//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "InjRea.h"

ClassImp(InjRea)

/////////////////////////////////////////////////////////////////////////
InjRea::InjRea(const string aPattern, const int aVerbose){ 
/////////////////////////////////////////////////////////////////////////

  // parameters
  Verbose=aVerbose;
  
  // init
  status_OK=true;
 
  // get injection tree
  // FIXME: this check sucks
  InjTree = new Chain("Gauss", aPattern.c_str(), Verbose); // try Gauss type
  if(!InjTree->GetEntries()){// try SineGauss type
    delete InjTree;
    InjTree = new Chain("SineGauss", aPattern.c_str(), Verbose);
  }
  if(!InjTree->GetEntries()){
    cerr<<"InjRea::InjRea: no injection found"<<endl;
    status_OK=false;
  }

  // attach branches
  if(Verbose) cout<<"InjRea::InjRea: connect "<<GetInjectionType()<<" TTree"<<endl;
  inj_ra=-1.0; inj_dec=-1.0; inj_time=-1.0; inj_ecc=-1.0; inj_psi=-1.0; inj_amp=-1.0; inj_sigma=-1.0; 
  InjTree->SetBranchAddress("ra",           &inj_ra);
  InjTree->SetBranchAddress("dec",          &inj_dec);
  InjTree->SetBranchAddress("time",         &inj_time);
  InjTree->SetBranchAddress("eccentricity", &inj_ecc);
  InjTree->SetBranchAddress("polarization", &inj_psi);
  InjTree->SetBranchAddress("amplitude",    &inj_amp);
  InjTree->SetBranchAddress("sigma",        &inj_sigma);
  if(!GetInjectionType().compare("SineGauss"))
    InjTree->SetBranchAddress("frequency",        &inj_f0);
  else inj_f0=0.0;// do not change missing value!

  inj_timefirst=0;
  inj_timelast=0;

  // check that injections are time sorted and with no overlap
  int ntrees=InjTree->GetNtrees();
  double te=0.0;
  for(int i=0; i<ntrees; i++){
    InjTree->GetEntry(InjTree->GetTreeOffset()[i]);// load first injection
    if(!i) inj_timefirst=inj_time;
    if(inj_time<te){
      cerr<<"InjRea::InjRea: the injection files must be time sorted and with no overlap"<<endl;
      status_OK=false;
      break;
    }
    InjTree->GetEntry(InjTree->GetTree()->GetEntries()-1);// load last injection
    inj_timelast=inj_time;
    te=inj_time;
  }

  // get min/max
  inj_sigmamin=InjTree->GetMinimum("sigma");
  inj_sigmamax=InjTree->GetMaximum("sigma");
  inj_ampmin=InjTree->GetMinimum("amplitude");
  inj_ampmax=InjTree->GetMaximum("amplitude");
  if(!GetInjectionType().compare("SineGauss")){
    inj_f0min=InjTree->GetMinimum("frequency");
    inj_f0max=InjTree->GetMaximum("frequency");
  }

  // tags
  inj_tag = new bool [InjTree->GetEntries()];
  for(int i=0; i<InjTree->GetEntries(); i++) inj_tag[i]=true;

  // set to first injection          
  if(InjTree->GetEntries()) InjTree->GetEntry(0);
  InjIndex=0;

  if(status_OK&&Verbose){
    cout<<"InjRea::InjRea: "<<InjTree->GetEntries()<<" "<<GetInjectionType()<<" injections are loaded"<<endl;
    if(Verbose){
      cout<<"                frequency: "<<inj_f0min<<" --> "<<inj_f0max<<" Hz"<<endl;
      cout<<"                sigma:     "<<inj_sigmamin<<" --> "<<inj_sigmamax<<" s"<<endl;
      cout<<"                amplitude: "<<scientific<<inj_ampmin<<" --> "<<inj_ampmax<<endl;
    }
  }
}

/////////////////////////////////////////////////////////////////////////
InjRea::~InjRea(void){
/////////////////////////////////////////////////////////////////////////
  if(Verbose>1) cout<<"InjRea::~InjRea"<<endl;
  delete InjTree;
  delete inj_tag;
}

/////////////////////////////////////////////////////////////////////////
double InjRea::GetSineGaussh0plus(void){
/////////////////////////////////////////////////////////////////////////
  // !!! here, amplitude = hrss !!!
  double q = 2 * TMath::Pi() * inj_sigma * inj_f0;
  
  // semimajor and semiminor axes of waveform ellipsoid
  const double a = 1.0 / sqrt(2.0 - inj_ecc * inj_ecc);
  const double b = a * sqrt(1.0 - inj_ecc * inj_ecc);

  // rss of plus polarization
  const double hplusrss  = inj_amp * (a * cos(inj_psi) - b * sin(inj_psi));
  
  // rss of unit amplitude cosine-gaussian waveforms.  see
  // K. Riles, LIGO-T040055-00.pdf
  const double cgrss = sqrt((q / (4.0 * inj_f0 * sqrt(TMath::Pi()))) * (1.0 + exp(-q * q)));
  
  // "peak" amplitude
  return hplusrss / cgrss;
}

/////////////////////////////////////////////////////////////////////////
double InjRea::GetSineGaussh0cross(void){
/////////////////////////////////////////////////////////////////////////
  // !!! here, amplitude = hrss !!!
  double q = 2 * TMath::Pi() * inj_sigma * inj_f0;

  // semimajor and semiminor axes of waveform ellipsoid
  const double a = 1.0 / sqrt(2.0 - inj_ecc * inj_ecc);
  const double b = a * sqrt(1.0 - inj_ecc * inj_ecc);
  
  // rss of cross polarizations
  const double hcrossrss = inj_amp * (b * cos(inj_psi) + a * sin(inj_psi));
  
  // rss of unit amplitude sine-gaussian waveforms.  see
  // K. Riles, LIGO-T040055-00.pdf
  const double sgrss = sqrt((q / (4.0 * inj_f0 * sqrt(TMath::Pi()))) * (1.0 - exp(-q * q)));

  // "peak" amplitude
  return hcrossrss / sgrss;
}


/////////////////////////////////////////////////////////////////////////
int InjRea::GetFirstInjectionIndex(const double aTimeStart, const double aTimeEnd){
/////////////////////////////////////////////////////////////////////////

  // out of range
  if(aTimeEnd<=inj_timefirst) return -1;
  if(aTimeStart>inj_timelast) return -1;

  if(inj_time>=aTimeStart){
    while(inj_time>=aTimeStart){// move backward
      if(InjIndex==0) return InjIndex;
      InjIndex--;
      InjTree->GetEntry(InjIndex);
      if(inj_time<aTimeStart){//went to far
	InjIndex++;
	InjTree->GetEntry(InjIndex);
	if(inj_time<aTimeEnd) return InjIndex;
	break;
      }
    }
  }

  else{
    while(InjIndex<GetNInjections()){// move foreward
      InjIndex++;
      InjTree->GetEntry(InjIndex);
      if(inj_time>=aTimeStart){
	if(inj_time<aTimeEnd) return InjIndex;
	break;
      }
    }
  }
    
  // not found
  return -1;
}

/////////////////////////////////////////////////////////////////////////
int InjRea::GetLastInjectionIndex(const double aTimeStart, const double aTimeEnd){
/////////////////////////////////////////////////////////////////////////

  // out of range
  if(aTimeEnd<=inj_timefirst) return -1;
  if(aTimeStart>inj_timelast) return -1;
  
  if(inj_time<aTimeEnd){
    while(inj_time<aTimeEnd){// move foreward
      if(InjIndex==GetNInjections()-1) return InjIndex;
      InjIndex++;
      InjTree->GetEntry(InjIndex);
      if(inj_time>=aTimeEnd){//went to far
	InjIndex--;
	InjTree->GetEntry(InjIndex);
	if(inj_time>=aTimeStart) return InjIndex;
	break;
      }
    }
  }

  else{
    while(InjIndex>=0){// move backward
      InjIndex--;
      InjTree->GetEntry(InjIndex);
      if(inj_time<aTimeEnd){
	if(inj_time>=aTimeStart) return InjIndex;
	break;
      }
    }
  }

    
  // not found
  return -1;
}

/////////////////////////////////////////////////////////////////////////
TH1D* InjRea::GetInjectionParamDist(const string aParamName, const int aNbins, const string aBinType){
/////////////////////////////////////////////////////////////////////////
  if(aNbins<1) return NULL;
  if(!InjTree->GetEntries()) return NULL;
  TH1D *h = NULL;

  // linear bins
  if(!aBinType.compare("UNIFORM")){
    if(!aParamName.compare("amplitude"))
      h = new TH1D("h_inj_amp","Injection amplitude",aNbins,inj_ampmin,TMath::Max(inj_ampmax,inj_ampmin*1.01));
    if(!aParamName.compare("f0"))
      h = new TH1D("h_inj_f0","Injection frequency",aNbins,inj_f0min,TMath::Max(inj_f0max,inj_f0min*1.01));
  }

  // log bins
  if(!aBinType.compare("LOG")){
    double *bins = new double [aNbins+1];
      
    if(!aParamName.compare("amplitude")){
      for(int b=0; b<aNbins+1; b++) 
	bins[b] = inj_ampmin*pow(10.0,b*log10(2*inj_ampmax/inj_ampmin)/(double)aNbins);
      h = new TH1D("h_inj_amp","Injection amplitude",aNbins,bins);
    }

    if(!aParamName.compare("f0")){
      for(int b=0; b<aNbins+1; b++) 
	bins[b] = inj_f0min*pow(10.0,b*log10(2*inj_f0max/inj_f0min)/(double)aNbins);
      h = new TH1D("h_inj_f0","Injection frequency",aNbins,bins);
    }

    delete bins;
  }

  // fill histo
  if(h!=NULL){
    h->GetYaxis()->SetTitle("Number of injections / bin");
    if(!aParamName.compare("amplitude")){
      h->GetXaxis()->SetTitle("Injection amplitude");
      for(int i=0; i<GetNInjections(); i++){
	LoadInjection(i);
	if(GetInjectionTag()) h->Fill(GetInjectionAmplitude());
      }
    }
    if(!aParamName.compare("f0")){
      h->GetXaxis()->SetTitle("Injection frequency [Hz]");
      for(int i=0; i<GetNInjections(); i++){
	LoadInjection(i);
	if(GetInjectionTag()) h->Fill(GetInjectionFrequency());
      }
    }
  } 
  
  return h;
}

