//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "Sample.h"

ClassImp(Sample)


////////////////////////////////////////////////////////////////////////////////////
Sample::Sample(const int aVerbosity){
////////////////////////////////////////////////////////////////////////////////////

  // parameters
  fNativeFrequency   = -1;
  fWorkingFrequency  = -1;// no resampling
  fHighPassFrequency = 0.0; // no highpassing
  fTukeyFraction     = 0.0;// no tukey window
  fRemoveDC          = false; // no DC removal
  fVerbosity         = aVerbosity;

  // nulling filters
  for(int i=0; i<NORDER; i++)// downsampling
    iirFilter[i]=NULL;
  
  for(int i=0; i<NORDERHP; i++)// highpass
    iirFilterHP[i]=NULL;
  
}

////////////////////////////////////////////////////////////////////////////////////
Sample::~Sample(){
////////////////////////////////////////////////////////////////////////////////////
  if(fVerbosity>1) cout<<"Sample::~Sample"<<endl;
  for(int i=0; i<NORDER; i++)   if(iirFilter[i]!=NULL)   delete iirFilter[i];
  for(int i=0; i<NORDERHP; i++) if(iirFilterHP[i]!=NULL) delete iirFilterHP[i];
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::Transform(const int invectorsize, double *invector, const int outvectorsize, double *outvector){
////////////////////////////////////////////////////////////////////////////////////
  if(fNativeFrequency<=0){
    cerr<<"Sample::Transform: frequencies must be set first. Use SetFrequencies()"<<endl;
    return false;
  }

  // check size consistency
  int induration  = invectorsize/fNativeFrequency;
  int outduration = outvectorsize/fWorkingFrequency;
  if((double)induration!=(double)invectorsize/(double)fNativeFrequency){
    cerr<<"Sample::Transform: the input vector must have a integer duration"<<endl;
    return false;
  }
  if((double)outduration!=(double)outvectorsize/(double)fWorkingFrequency){
    cerr<<"Sample::Transform: the output vector must have a integer duration"<<endl;
    return false;
  }
  if(outduration!=induration){
    cerr<<"Sample::Transform: the input and output vectors must have the same duration"<<endl;
    return false;
  }

  // requested DC removal
  if(fRemoveDC){
    if(fVerbosity) cout<<"Sample::Transform: remove DC..."<<endl;
    RemoveDC(invectorsize,invector);
  }

  // requested highpass filtering
  if(fHighPassFrequency>0.0){
    if(fVerbosity) cout<<"Sample::Transform: highpass input data..."<<endl;
    if(!HighPass(invectorsize, invector)) return false;
  }

  // requested windowing
  if(fTukeyFraction){
    if(fVerbosity) cout<<"Sample::Transform: window data..."<<endl;
    if(!ApplyTukeyWindow(invectorsize,invector)) return false;
  }

  // requested resampling
  if(fWorkingFrequency!=fNativeFrequency){
    if(fVerbosity) cout<<"Sample::Transform: resample input data..."<<endl;
    if(!Resample(invectorsize, invector, outvector)) return false;
  }
  else// make a copy
    for(int i=0; i<invectorsize; i++) outvector[i]=invector[i];

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::SetWindows(const double aTukeyFraction){
////////////////////////////////////////////////////////////////////////////////////

  if(aTukeyFraction<0.0 || aTukeyFraction>1.0){
    cerr<<"Sample::SetWindows: the Tukey window fraction must be between 0 and 1"<<endl;
    return false;
  }
  fTukeyFraction=aTukeyFraction;
  if(fVerbosity) cout<<"Sample::SetWindows: new Tukey window fraction ("<<fTukeyFraction<<")"<<endl;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::SetFrequencies(const int aNativeFrequency, const int aWorkingFrequency, const double aHighPassFrequency){
////////////////////////////////////////////////////////////////////////////////////
  
  // native
  if(aNativeFrequency<=0){
    cerr<<"Sample::SetFrequencies: the native sampling frequency is not valid"<<endl;
    return false;
  }

  // working
  if(aWorkingFrequency<=0){
    cerr<<"Sample::SetFrequencies: the working sampling frequency is not valid"<<endl;
    return false;
  }
  if(aWorkingFrequency>aNativeFrequency){
    cerr<<"Sample::SetFrequencies: upsampling is not supported"<<endl;
    return false;
  }
  if (!IsPowerOfTwo(aWorkingFrequency)){
    cerr<<"Sample::SetFrequencies: the working sampling frequency can only be a power of 2"<<endl;
    return false;
  }

  // highpass
  if(aHighPassFrequency<0.0){
    cerr<<"Sample::SetFrequencies: the highpass frequency is not valid"<<endl;
    return false;
  }
  if(aHighPassFrequency>=aWorkingFrequency){
    cerr<<"Sample::SetFrequencies: the highpass frequency is larger than the working sampling frequency"<<endl;
    return false;
  }

  // update filters
  if(aNativeFrequency!=fNativeFrequency){
    if(fVerbosity) cout<<"Sample::SetFrequencies: new native sampling frequency ("<<aNativeFrequency<<"Hz)"<<endl;
    fNativeFrequency  = aNativeFrequency;

    if(aWorkingFrequency!=fWorkingFrequency){
      if(fVerbosity) cout<<"Sample::SetFrequencies: new working sampling frequency ("<<aWorkingFrequency<<"Hz)"<<endl;
      fWorkingFrequency = aWorkingFrequency;

      if(aHighPassFrequency!=fHighPassFrequency){
	if(fVerbosity) cout<<"Sample::SetFrequencies: new highpass frequency ("<<aHighPassFrequency<<"Hz)"<<endl;
	fHighPassFrequency = aHighPassFrequency;
      }
      return MakeSamplingFilters()&&MakeHighPassFilters();
    }

    return MakeSamplingFilters();
  }

  if(aWorkingFrequency!=fWorkingFrequency){
    if(fVerbosity) cout<<"Sample::SetFrequencies: new working sampling frequency ("<<aWorkingFrequency<<"Hz)"<<endl;
    fWorkingFrequency = aWorkingFrequency;
    
    if(aHighPassFrequency!=fHighPassFrequency){
      if(fVerbosity) cout<<"Sample::SetFrequencies: new highpass frequency ("<<aHighPassFrequency<<"Hz)"<<endl;
      fHighPassFrequency = aHighPassFrequency;
    }
    return MakeSamplingFilters()&&MakeHighPassFilters();
  }

  if(aHighPassFrequency!=fHighPassFrequency){
    if(fVerbosity) cout<<"Sample::SetFrequencies: new highpass frequency ("<<aHighPassFrequency<<"Hz)"<<endl;
    fHighPassFrequency = aHighPassFrequency;
    return MakeHighPassFilters();
  }

  // nothing changed, so nothing to do!
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::SetNativeFrequency(const int aNativeFrequency){
////////////////////////////////////////////////////////////////////////////////////
  return SetFrequencies(aNativeFrequency,fWorkingFrequency,fHighPassFrequency);
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::SetWorkingFrequency(const int aWorkingFrequency){
////////////////////////////////////////////////////////////////////////////////////
  return SetFrequencies(fNativeFrequency,aWorkingFrequency,fHighPassFrequency);
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::SetHighPassFrequency(const double aHighPassFrequency){
////////////////////////////////////////////////////////////////////////////////////
  return SetFrequencies(fNativeFrequency,fWorkingFrequency,aHighPassFrequency);
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::MakeSamplingFilters(void){
////////////////////////////////////////////////////////////////////////////////////

  // presampling is required
  if(!IsPowerOfTwo(fNativeFrequency)){
    presample=true;
    double powerof2=floor(log(fNativeFrequency)/log(2.0));
    fIntermediateFrequency=(int)pow(2.0,powerof2);
    if(fVerbosity>1) cout<<"Sample::MakeSamplingFilters: presampling is required ("<<fIntermediateFrequency<<"Hz)"<<endl;
    fResampleFactor = fIntermediateFrequency / fWorkingFrequency;
  }
  else{
    presample=false;
    fIntermediateFrequency=fNativeFrequency;
    fResampleFactor = fNativeFrequency / fWorkingFrequency;
  }
   
  // locals
  double w1;
  double a1;
  double wc;
  double theta,ar,ai;
  
  // build downsampling filters
  ZPGFilter *zpgFilter[NORDER];
  for(int i=0; i<NORDER; i++)
    if(iirFilter[i]!=NULL){ delete iirFilter[i]; iirFilter[i]=NULL; }

  // Filter parameters
  w1=(double)fWorkingFrequency/2.0/(double)fIntermediateFrequency;// Dimensionless frequencies;
  a1=0.1;// Attenuation factor;
  wc=TMath::Tan(TMath::Pi()*w1)*pow(1.0/sqrt(a1) - 1.0, -0.5/((double)(NORDER)));
  
  // An order n Butterworth filter has n poles spaced evenly along a
  // semicircle in the upper complex w-plane.  By pairing up poles
  // symmetric across the imaginary axis, the filter can be decomposed
  // into [n/2] filters of order 2, plus perhaps an additional order 1
  // filter.  The following loop pairs up poles and applies the
  // filters with order 2
  if(fVerbosity>1) cout<<"Sample::MakeSamplingFilters: building butterworth filters"<<endl;
  for(int i=0,j=NORDER-1; i<j; i++,j--){
    theta=TMath::Pi()*(i+0.5)/(double)NORDER;
    ar=wc*TMath::Cos(theta);
    ai=wc*TMath::Sin(theta);
    zpgFilter[i]=new ZPGFilter(0,2);
    zpgFilter[i]->gain[0]=-wc*wc;
    zpgFilter[i]->gain[1]=0.0;
    zpgFilter[i]->poles[0][0]=ar;
    zpgFilter[i]->poles[1][0]=ai;
    zpgFilter[i]->poles[0][1]=-ar;
    zpgFilter[i]->poles[1][1]=ai;
    
    // Transform to the z-plane
    if(!zpgFilter[i]->WToZ()){
      cerr<<"Sample::MakeSamplingFilters: transform to the z-plane failed"<<endl;
      return false;
    }  
    
    iirFilter[i] = new IIRFilter(zpgFilter[i]);
    delete zpgFilter[i];
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::MakeHighPassFilters(void){
////////////////////////////////////////////////////////////////////////////////////

  // locals
  double w1;
  double a1;
  double wc;
  double theta,ar,ai;
  
  // build high-pass filters
  ZPGFilter *zpgFilterHP[NORDERHP];
  for(int i=0; i<NORDERHP; i++)
    if(iirFilterHP[i]!=NULL){ delete iirFilterHP[i]; iirFilterHP[i]=NULL; }

  // Filter parameters
  w1=fHighPassFrequency/(double)fNativeFrequency;// Dimensionless frequency;
  a1=0.98;// Attenuation factor;
  wc=TMath::Tan(TMath::Pi()*w1)*pow(1.0/sqrt(a1) - 1.0, 0.5/((double)(NORDERHP)));
  
  if(fVerbosity>1) cout<<"Sample::SetHighPassFrequency: building high-pass filters"<<endl;
  for(int i=0,j=NORDERHP-1; i<j; i++,j--){
    theta=TMath::Pi()*(i+0.5)/(double)NORDERHP;
    ar=wc*TMath::Cos(theta);
    ai=wc*TMath::Sin(theta);
    zpgFilterHP[i]=new ZPGFilter(2,2);
    zpgFilterHP[i]->gain[0]=1.0;
    zpgFilterHP[i]->gain[1]=0.0;
    zpgFilterHP[i]->zeros[0][0]=0.0;
    zpgFilterHP[i]->zeros[1][0]=0.0;
    zpgFilterHP[i]->zeros[0][1]=0.0;
    zpgFilterHP[i]->zeros[1][1]=0.0;
    zpgFilterHP[i]->poles[0][0]=ar;
    zpgFilterHP[i]->poles[1][0]=ai;
    zpgFilterHP[i]->poles[0][1]=-ar;
    zpgFilterHP[i]->poles[1][1]=ai;
      
    // Transform to the z-plane
    if(!zpgFilterHP[i]->WToZ()){
      cerr<<"Sample::SetHighPassFrequency: transform to the z-plane failed"<<endl;
      return false;
    }  
      
    iirFilterHP[i] = new IIRFilter(zpgFilterHP[i]);
    delete zpgFilterHP[i];
  }
  
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::Resample(int invectorsize, double *invector, double *outvector){
////////////////////////////////////////////////////////////////////////////////////

  int duration = invectorsize/fNativeFrequency;

  // presampling required
  if(presample){
    if(fVerbosity) cout<<"Sample::Resample: apply pre-sampling"<<endl;
    if(!PreSample(invectorsize, invector)){
      cerr<<"Sample::Resample: pre-sampling failed"<<endl;
      return false;
    }
  }

  // apply filter
  if(fResampleFactor>1){
    if(fVerbosity) cout<<"Sample::Resample: apply anti-aliasing filters"<<endl;
    for(int ii=0,jj=NORDER-1; ii<jj; ii++,jj--)
      if(!ApplyFilter(invectorsize, invector, iirFilter[ii])){
	cerr<<"Sample::Resample: filtering failed"<<endl;
	return false;
      }
  }

  //decimate time series
  if(fVerbosity) cout<<"Sample::Resample: decimate time-series"<<endl;
  int outsize = duration*fWorkingFrequency;
  for (int j=0; j<outsize; ++j){
    outvector[j] = *invector;
    invector += fResampleFactor;
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::HighPass(const int invectorsize, double *invector){
////////////////////////////////////////////////////////////////////////////////////

  // apply filter
  if(fVerbosity) cout<<"Sample::HighPass: apply filters"<<endl;
  for(int i=0,j=NORDERHP-1; i<j; i++,j--){
    if(!ApplyFilter(invectorsize, invector, iirFilterHP[i])){
      cerr<<"Sample::HighPass: filtering failed"<<endl;
      return false;
    }
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::PreSample(int &datavectorsize, double *datavector){
////////////////////////////////////////////////////////////////////////////////////
      
  // input vector duration
  int duration   = datavectorsize/fNativeFrequency;
  int newsize    = duration*fIntermediateFrequency;

  double si,ei,sj,ej,newvalue,wsum;
  int istart=0;
  for (int i=0; i<newsize; i++){// NEW vector
    si=(double)i/(double)fIntermediateFrequency;// bin start
    ei=(double)(i+1)/(double)fIntermediateFrequency;// bin end
    newvalue=0;
    wsum=0;
    for (int j=istart; j<datavectorsize; j++){// OLD vector
      sj=(double)j/(double)fNativeFrequency;// bin start
      ej=(double)(j+1)/(double)fNativeFrequency;// bin end
      if(ej<si){ continue; }
      else if(sj<si&&ej>=si){ newvalue+=((ej-si)*datavector[j]); wsum+=(ej-si); }
      else if(sj>=si&&ej<ei){ newvalue+=((ej-sj)*datavector[j]); wsum+=(ej-sj); }
      else if(sj<ei&&ej>=ei){ newvalue+=((ei-sj)*datavector[j]); wsum+=(ei-sj); }
      else{ istart=j-1; break; }
    }
    if(wsum) newvalue/=wsum;
    else{
      cerr<<"SHOULD NEVER HAPPEN!!!"<<endl;
      return false;
    }

    //cout<<i<<endl;
    datavector[i] = newvalue;
    //cout<<i<<endl;
  }

  // set new size
  datavectorsize=newsize;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::ApplyFilter(const int datavectorsize, double *datavector, IIRFilter *filter){
////////////////////////////////////////////////////////////////////////////////////
  
  int directOrder = filter->numDirect;
  int recursOrder = filter->numRecurs;
  int numHist     = filter->numRecurs;
  double *temp    = new double [numHist];
  double datum;
  int i,j,k;
   
  // Compute the auxiliary data series
  for(i=0;(i<recursOrder)&&(i<datavectorsize);i++,datavector++){
    datum=*datavector;
    for(j=1;j<=i;j++)
      datum+=datavector[-j]*filter->recursCoef[j];
    for(k=0;j<recursOrder;j++,k++)
      datum+=filter->history[k]*filter->recursCoef[j];
    *datavector=datum;
  }
  
  for(;i<datavectorsize;i++,datavector++){
    datum=*datavector;
    for(j=1;j<recursOrder;j++)
      datum+=datavector[-j]*filter->recursCoef[j];
    *datavector=datum;
  }
  datavector--;
  
  // Store the last few auxiliary data to the temporary history
  for(k=numHist-1;k>=datavectorsize;k--)
    temp[k]=filter->history[k-datavectorsize];
  for(;k>=0;k--)
    temp[k]=datavector[-k];
  
  // Compute the output data series
  for(;i>directOrder;i--,datavector--){
    datum=*datavector*filter->directCoef[0];
    for(j=1;j<directOrder;j++)
      datum+=datavector[-j]*filter->directCoef[j];
    *datavector=datum;
  }
  for(;i>0;i--,datavector--){
    datum=*datavector*filter->directCoef[0];
    for(j=1;j<i;j++)
      datum+=datavector[-j]*filter->directCoef[j];
    for(k=0;j<directOrder;j++,k++)
      datum+=filter->history[k]*filter->directCoef[j];
    *datavector=datum;
  }
  
  // Update the filter history from the temporary history
  for(k=0;k<numHist;k++)
    filter->history[k]=temp[k];
  delete temp;
  
  // prepare for reverse
  datavector+=datavectorsize;
  
  // Perform the auxilliary piece of the filter
  for(i=0;(i<recursOrder)&&(i<datavectorsize);i++,datavector--){
    datum=*datavector;
    for(j=1;j<=i;j++)
      datum+=datavector[j]*filter->recursCoef[j];
    *datavector=datum;
  }
  
  for(;i<datavectorsize;i++,datavector--){
    datum=*datavector;
    for(j=1;j<recursOrder;j++)
      datum+=datavector[j]*filter->recursCoef[j];
    *datavector=datum;
  }
  
  datavector++;
  
  // Perform the direct piece of the filter
  for(;i>directOrder;i--,datavector++){
    datum=*datavector*filter->directCoef[0];
    for(j=1;j<directOrder;j++)
      datum+=datavector[j]*filter->directCoef[j];
    *datavector=datum;
  }
  for(;i>0;i--,datavector++){
    datum=*datavector*filter->directCoef[0];
    for(j=1;j<i;j++)
      datum+=datavector[j]*filter->directCoef[j];
    *datavector=datum;
  }
  
  // reset history
  for(int i=0; i<numHist; i++) filter->history[i]=filter->history_init[i];
  
  // back to begining
  datavector-=datavectorsize;
  
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
void Sample::RemoveDC(const int datavectorsize, double *datavector){
////////////////////////////////////////////////////////////////////////////////////
  if(!fRemoveDC) return; // nothing to do

  double dc=0.0;
  for (int i=0; i<datavectorsize; i++) dc+=datavector[i];
  dc/=datavectorsize;
  for (int i=0; i<datavectorsize; i++) datavector[i] -= dc;
 
  return;
}

////////////////////////////////////////////////////////////////////////////////////
bool Sample::ApplyTukeyWindow(const int datavectorsize, double *datavector){
////////////////////////////////////////////////////////////////////////////////////
  if(!fTukeyFraction) return true; // nothing to do
  
  int FracSize_2=(int)floor(fTukeyFraction*(double)datavectorsize/2.0);
    
  double factor;
  if(FracSize_2) factor = TMath::Pi()/(double)FracSize_2;
  else return false;// should never happen

  // beginning
  for (int i=0; i<FracSize_2; i++)
    datavector[i] *= 0.5*(1+TMath::Cos(factor*(double)(i-FracSize_2)));

  // end
  for (int i=datavectorsize-FracSize_2; i<datavectorsize; i++)
    datavector[i] *= 0.5*(1+TMath::Cos(factor*(double)(i-datavectorsize+FracSize_2)));
 
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
ZPGFilter::ZPGFilter(const unsigned int numZeros, const unsigned int numPoles){ 
////////////////////////////////////////////////////////////////////////////////////
 
  // gain
  gain[0]=0;
  gain[1]=0;
    
  // zeros
  nzeros=numZeros;
  zeros[0] = new double [nzeros];
  zeros[1] = new double [nzeros];
  
  // poles
  npoles=numPoles;
  poles[0] = new double [npoles];
  poles[1] = new double [npoles];

}

////////////////////////////////////////////////////////////////////////////////////
ZPGFilter::~ZPGFilter(void){ 
////////////////////////////////////////////////////////////////////////////////////
  delete zeros[0];
  delete zeros[1];
  delete poles[0];
  delete poles[1];
}

////////////////////////////////////////////////////////////////////////////////////
bool ZPGFilter::WToZ(void){ 
////////////////////////////////////////////////////////////////////////////////////

  // locals
  double *g0, *g1, *b0, *b1;

  // Compute the total number of zeros and poles in the w-plane,
  // including those at w=infinity
  unsigned int numPoles=npoles;
  unsigned int numZeros=nzeros;
  unsigned int num = (numZeros>numPoles) ? numZeros : numPoles;
  numZeros=numPoles=num;

  // Compute the revised number of zeros and poles in the z-plane,
  // excluding those at z=infinity (w=-i)
  for(unsigned int i=0; i<npoles; i++)
    if(poles[0][i]==0.0&&poles[1][i]==-1.0) numPoles--;
  for(unsigned int i=0; i<nzeros; i++)
    if(zeros[0][i]==0.0&&zeros[1][i]==-1.0) numZeros--;

  // new gains
  double *newgain[2];
  unsigned int ng = npoles+nzeros;
  newgain[0] = new double [ng];
  newgain[1] = new double [ng];

  // new zeros
  double *newzeros[2];
  newzeros[0] = new double [numZeros];
  newzeros[1] = new double [numZeros];

  // new poles
  double *newpoles[2];
  newpoles[0] = new double [numPoles];
  newpoles[1] = new double [numPoles];

  // Transform existing zeros from w to z, except for those at w=-i,
  // which are mapped to z=infinity.  At the same time, compute the
  // gain correction factors.
  g0=newgain[0];
  g1=newgain[1];
  b0=newzeros[0];
  b1=newzeros[1];
  double zr, zi, ratio, denom;
  unsigned int iz=0, ig=0;
  for(unsigned int i=0; i<nzeros; i++,ig++){
    zr=zeros[0][i];
    zi=zeros[1][i];
    if(zr==0.0){
      if(zi==-1.0){// w=-i is mapped to z=infinity
	g0[ig]=0.0;
	g1[ig]=2.0;
      }
      else{// w=i*y is mapped to z=(1-y)/(1+y)
	b0[iz]=(1.0-zi)/(1.0+zi);
	b1[iz]=0.0;
	g0[ig]=0.0;
	g1[ig]=-(1.0+zi);
	iz++;
      }
    }
    else if(fabs(1.0+zi)>fabs(zr)){
      ratio = -zr/(1.0+zi);
      denom = 1.0+zi - ratio*zr;
      b0[iz] = (1.0-zi + ratio*zr)/denom;
      b1[iz] = (zr - ratio*(1.0-zi))/denom;
      g0[ig] = -zr;
      g1[ig] = -(1.0+zi);
      iz++;
    }
    else{
      ratio = -(1.0+zi)/zr;
      denom = -zr + ratio*(1.0+zi);
      b0[iz] = ((1.0-zi)*ratio + zr)/denom;
      b1[iz] = (zr*ratio - 1.0+zi)/denom;
      g0[ig] = -zr;
      g1[ig] = -(1.0+zi);
      iz++;
    }
  }
  // Transform any remaining zeros at w=infinity to z=-1
  for(; iz<numZeros; iz++){
    b0[iz] = -1.0;
    b1[iz] = 0.0;
  }
  // Replace the old filter zeros with the new ones
  if(nzeros>0){
    delete zeros[0];
    delete zeros[1];
  }
  zeros[0] = b0;
  zeros[1] = b1;
  nzeros=numZeros;

  // Transform existing poles from w to z, except for those at w=-i,
  // which are mapped to z=infinity.  At the same time, compute the
  // gain correction factors
  b0=newpoles[0];
  b1=newpoles[1];
  double pr, pi;
  unsigned int ip=0;
  for(unsigned int i=0; i<npoles; i++,ig++){
    pr=poles[0][i];
    pi=poles[1][i];
    if(pr==0.0){
      if(pi==-1.0){// w=-i is mapped to z=infinity
	g0[ig]=0.0;
	g1[ig]=-0.5;
      }
      else{// w=i*y is mapped to z=(1-y)/(1+y)
	b0[ip]=(1.0-pi)/(1.0+pi);
	b1[ip]=0.0;
	g0[ig]=0.0;
	g1[ig]=1.0/(1.0+pi);
	ip++;
      }
    }
    else if(fabs(1.0+pi)>fabs(pr)){
      ratio = -pr/(1.0+pi);
      denom = 1.0+pi - ratio*pr;
      b0[ip] = (1.0-pi + ratio*pr)/denom;
      b1[ip] = (pr - ratio*(1.0-pi))/denom;
      g0[ig] = ratio/denom;
      g1[ig] = 1.0/denom;
      ip++;
    }
    else{
      ratio = -(1.0+pi)/pr;
      denom = -pr + ratio*(1.0+pi);
      b0[ip] = ((1.0-pi)*ratio + pr)/denom;
      b1[ip] = (pr*ratio - 1.0+pi)/denom;
      g0[ig] = ratio/denom;
      g1[ig] = 1.0/denom;
      ip++;
    }
  }
  // Transform any remaining poles at w=infinity to z=-1
  for(; ip<numPoles; ip++){
    b0[ip] = -1.0;
    b1[ip] = 0.0;
  }
  // Replace the old filter poles with the new ones
  if(npoles>0){
    delete poles[0];
    delete poles[1];
  }
  poles[0] = b0;
  poles[1] = b1;
  npoles=numPoles;

  // FIXME: the gains should be multiplied alternately by small and large correction factors
  // Multiply the small and largest factors together
  unsigned int i=0,j=ng-1;
  for(;i<j;i++,j--){
    double ar=g0[i];
    double ai=g1[i];
    double br=g0[j];
    double bi=g1[j];
    double cr=ar*br-ai*bi;
    double ci=ar*bi+ai*br;
  
    // Multiply the gain by the combined factor
    br=gain[0];
    bi=gain[1];
    gain[0]=br*cr-bi*ci;
    gain[1]=br*ci+bi*cr;
  }
  if(i==j){
    /* Multiply by the remaining odd factor. */
    double cr=g0[i];
    double ci=g1[i];
    double br=gain[0];
    double bi=gain[1];

    gain[0]=br*cr-bi*ci;
    gain[1]=br*ci+bi*cr;
  }

  delete newgain[0];
  delete newgain[1];

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
IIRFilter::IIRFilter(ZPGFilter *zpgFilter){ 
////////////////////////////////////////////////////////////////////////////////////
 
  nzeros=zpgFilter->nzeros;
  npoles=zpgFilter->npoles;
  zeros[0]=zpgFilter->zeros[0];
  zeros[1]=zpgFilter->zeros[1];
  poles[0]=zpgFilter->poles[0];
  poles[1]=zpgFilter->poles[1];
  unsigned int num=0;
  
  numDirect=1;
  for(unsigned int i=0; i<nzeros; i++)
    if(zeros[1][i]==0.0){
      num+=1;
      if(zeros[0][i]==0.0)
	numDirect-=1;
    }
    else if(zeros[1][i]>0.0){
      num+=2;
    }
  
  numRecurs=1; num=0;
  for(unsigned int i=0; i<npoles; i++)
    if(poles[1][i]==0.0){
      num+=1;
      if(poles[0][i]==0.0)
	numRecurs-=1;
    }
    else if(poles[1][i]>0.0){
      num+=2;
    }
  
  num = (npoles>=nzeros) ? nzeros :  npoles;
  numDirect+=num;
  numRecurs+=num;

  directCoef = new double [numDirect];
  recursCoef = new double [numRecurs];
    
  double *direct;// The direct filter coefficients
  double *recurs;// The recursive filter coefficients
  direct=directCoef;
  recurs=recursCoef;

  // Expand the denominator as a polynomial in z
  *recurs=-1.0;
  for(unsigned int i=1; i<numRecurs; i++) recurs[i]=0.0;
  for(unsigned int i=0; i<npoles; i++){
    unsigned int j=numRecurs-1;
    double x=poles[0][i];
    double y=poles[1][i];
    if(y==0.0)
      for(; j>0; j--) recurs[j]-=x*recurs[j-1];
    else if(y>0.0){
      for(; j>1; j--) recurs[j]-=2.0*x*recurs[j-1]-(x*x+y*y)*recurs[j-2];
      recurs[j]-=2.0*x*recurs[j-1];
    }
  }

  // Expand the numerator as a polynomial in z
  for(unsigned int i=0; i<numDirect; i++) direct[i]=0.0;
  direct[num-nzeros]=zpgFilter->gain[0];
  for(unsigned int i=0; i<nzeros; i++){
    unsigned int j=numDirect-1;
    double x=zeros[0][i];
    double y=zeros[1][i];
    if(y==0.0)
      for(; j>num-nzeros; j--) direct[j]-=x*direct[j-1];
    else if(y>0.0){
      for(; j>num-nzeros+1; j--) direct[j]-=2.0*x*direct[j-1]-(x*x+y*y)*direct[j-2];
      direct[j]-=2.0*x*direct[j-1];
    }
  }

  // Initialize filter history
  if(numDirect>=numRecurs)
    num=numDirect-1;
  else
    num=numRecurs-1;
  history = new double [numRecurs];
  history_init = new double [numRecurs];
  for(unsigned int i=0; i<num; i++) history[i]=0.0;
  for(unsigned int i=0; i<num; i++) history_init[i]=history[i];
 
}

  ////////////////////////////////////////////////////////////////////////////////////
IIRFilter::~IIRFilter(void){ 
////////////////////////////////////////////////////////////////////////////////////

  // CAUTION: poles and zeroes do not need to be freed 
  // since they are to be freed for the ZPGFilter structure given as input
  delete directCoef;
  delete recursCoef;
  delete history;
  delete history_init;

}
