//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "Streams.h"

ClassImp(Streams)

/////////////////////////////////////////////////////////////////////////
Streams::Streams(const string aName, const int aVerbose): Sample(aVerbose) { 
/////////////////////////////////////////////////////////////////////////

  // init
  verbose=aVerbose;
  
  // get name
  Name=aName;
  if(Name.length()<4){
    cerr<<"Streams::Streams: the stream name must have the following format: [prefix]:[suffix]"<<endl;
    cerr<<"                  --> set to default: 00:none"<<endl;
    Name="00:none";
  }

  // no prefix
  if((Name.substr(2,1)).compare(":")) Name="00:"+aName;
    
  // prefix/suffix
  NamePrefix=Name.substr(0,2);
  NameSuffix=Name.substr(3,Name.length()-3);
  Name=NamePrefix+":"+NameSuffix;
  NameSuffixUnderScore=ReplaceAll(NameSuffix,"-","_");

  // try to make a LV detector
  MakeLVDetector();
}

/////////////////////////////////////////////////////////////////////////
Streams::~Streams(void){
/////////////////////////////////////////////////////////////////////////
  if(verbose>1) cout<<"Streams::~Streams"<<endl;
}

/////////////////////////////////////////////////////////////////////////
bool Streams::SetName(const string aNewName){
/////////////////////////////////////////////////////////////////////////
 
  if(aNewName.length()<4){
    cerr<<"Streams::SetName: the stream name must have the following format: [prefix]:[suffix]"<<endl;
    return false;
  }
  Name=aNewName;
 
  // prefix/suffix
  NamePrefix=Name.substr(0,2);
  NameSuffix=Name.substr(3,Name.length()-3);
  Name=NamePrefix+":"+NameSuffix;
  NameSuffixUnderScore=ReplaceAll(NameSuffix,"-","_");

  if(verbose) cout<<"Streams::SetName: this stream has a new name: "<<Name<<endl;
  
  // try to make a LV detector
  MakeLVDetector();

  return true;
}

//***************************************************************************
// Below this line ---> detector applications
//***************************************************************************

/////////////////////////////////////////////////////////////////////////
void Streams::MakeLVDetector(void){
/////////////////////////////////////////////////////////////////////////

  // default
  DetIndex=-1;
  ComputeDetectorResponse(); // zero-out
  DetLoc[0] = 0.0;
  DetLoc[1] = 0.0;
  DetLoc[2] = 0.0;

  // get det index based on the prefix
  DetIndex=GetLVDetIndex(NamePrefix);

  // nice try!
  if(DetIndex<0){
    if(verbose) cout<<"Streams::MakeDetector: this stream does not match any known LV detector"<<endl;
    return;
  }
  if(verbose) cout<<"Streams::MakeLVDetector: this stream is now a LV detector: "<<NamePrefix<<" detindex="<<DetIndex<<endl;

  ComputeDetectorResponse();
  DetLoc[0] = DET_VERTEX_LOCATION_X_SI[DetIndex];
  DetLoc[1] = DET_VERTEX_LOCATION_Y_SI[DetIndex];
  DetLoc[2] = DET_VERTEX_LOCATION_Z_SI[DetIndex];

  return;
}

/////////////////////////////////////////////////////////////////////////
void Streams::GetDetAMResponse(double &fplus, double &fcross, const double ra, const double dec, const double psi, const double gmst){
/////////////////////////////////////////////////////////////////////////

  // An implementation of the detector response formulae in Anderson et al
  // PRD 63 042003 (2001)
  // Computes F+ and Fx for a source at a specified sky position,
  // polarization angle, and sidereal time.
  
  double X[3];
  double Y[3];

  // Greenwich hour angle of source (radians)
  const double gha = gmst - ra;

  // pre-compute trig functions
  const double cosgha = cos(gha);
  const double singha = sin(gha);
  const double cosdec = cos(dec);
  const double sindec = sin(dec);
  const double cospsi = cos(psi);
  const double sinpsi = sin(psi);

  // Note that dec = pi/2 - theta, and gha =
  // -phi where theta and phi are the standard spherical coordinates
  // used in that paper
  X[0] = -cospsi * singha - sinpsi * cosgha * sindec;
  X[1] = -cospsi * cosgha + sinpsi * singha * sindec;
  X[2] =  sinpsi * cosdec;
  Y[0] =  sinpsi * singha - cospsi * cosgha * sindec;
  Y[1] =  sinpsi * cosgha + cospsi * singha * sindec;
  Y[2] =  cospsi * cosdec;

  fplus = fcross = 0.0;
  double DX, DY;
  for(int i = 0; i < 3; i++){
    DX = Response[i][0] * X[0] + Response[i][1] * X[1] + Response[i][2] * X[2];
    DY = Response[i][0] * Y[0] + Response[i][1] * Y[1] + Response[i][2] * Y[2];
    fplus  += X[i] * DX - Y[i] * DY;
    fcross += X[i] * DY + Y[i] * DX;
  }

  return;
}

/////////////////////////////////////////////////////////////////////////
void Streams::ComputeDetectorResponse(void){
/////////////////////////////////////////////////////////////////////////
  if(DetIndex<0||DetIndex>N_KNOWN_DET){// bad det index
    Response[0][0]=0.0;
    Response[0][1]=0.0;
    Response[0][2]=0.0;
    Response[1][0]=0.0;
    Response[1][1]=0.0;
    Response[1][2]=0.0;
    Response[2][0]=0.0;
    Response[2][1]=0.0;
    Response[2][2]=0.0;
  }
  else{// OK
    Response[0][0]=0.5*(DET_ARM_X_DIRECTION_X[DetIndex]*DET_ARM_X_DIRECTION_X[DetIndex]-DET_ARM_Y_DIRECTION_X[DetIndex]*DET_ARM_Y_DIRECTION_X[DetIndex]);
    Response[0][1]=0.5*(DET_ARM_X_DIRECTION_X[DetIndex]*DET_ARM_X_DIRECTION_Y[DetIndex]-DET_ARM_Y_DIRECTION_X[DetIndex]*DET_ARM_Y_DIRECTION_Y[DetIndex]);
    Response[0][2]=0.5*(DET_ARM_X_DIRECTION_X[DetIndex]*DET_ARM_X_DIRECTION_Z[DetIndex]-DET_ARM_Y_DIRECTION_X[DetIndex]*DET_ARM_Y_DIRECTION_Z[DetIndex]);
    Response[1][0]=0.5*(DET_ARM_X_DIRECTION_Y[DetIndex]*DET_ARM_X_DIRECTION_X[DetIndex]-DET_ARM_Y_DIRECTION_Y[DetIndex]*DET_ARM_Y_DIRECTION_X[DetIndex]);
    Response[1][1]=0.5*(DET_ARM_X_DIRECTION_Y[DetIndex]*DET_ARM_X_DIRECTION_Y[DetIndex]-DET_ARM_Y_DIRECTION_Y[DetIndex]*DET_ARM_Y_DIRECTION_Y[DetIndex]);
    Response[1][2]=0.5*(DET_ARM_X_DIRECTION_Y[DetIndex]*DET_ARM_X_DIRECTION_Z[DetIndex]-DET_ARM_Y_DIRECTION_Y[DetIndex]*DET_ARM_Y_DIRECTION_Z[DetIndex]);
    Response[2][0]=0.5*(DET_ARM_X_DIRECTION_Z[DetIndex]*DET_ARM_X_DIRECTION_X[DetIndex]-DET_ARM_Y_DIRECTION_Z[DetIndex]*DET_ARM_Y_DIRECTION_X[DetIndex]);
    Response[2][1]=0.5*(DET_ARM_X_DIRECTION_Z[DetIndex]*DET_ARM_X_DIRECTION_Y[DetIndex]-DET_ARM_Y_DIRECTION_Z[DetIndex]*DET_ARM_Y_DIRECTION_Y[DetIndex]);
    Response[2][2]=0.5*(DET_ARM_X_DIRECTION_Z[DetIndex]*DET_ARM_X_DIRECTION_Z[DetIndex]-DET_ARM_Y_DIRECTION_Z[DetIndex]*DET_ARM_Y_DIRECTION_Z[DetIndex]);
  }
  return;
}
