//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __TriggerBuffer__
#define __TriggerBuffer__
#include "CUtils.h"
#include "MakeTriggers.h"

using namespace std;

/**
 * Create and manage a buffer for triggers.
 * See also MakeTriggers()
 *
 * \author    Florent Robinet
 */
class TriggerBuffer: public MakeTriggers {

 public:
  
  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the TriggerBuffer class.
   * The size of the buffer is defined. This size determines the maximum number of triggers which can be saved in the buffer. The memory for the buffer is then allocated.
   * @param aSize buffer size
   * @param aStreamName stream name
   * @param aVerbose verbosity level
   */
  TriggerBuffer(const int aSize, const string aStreamName, const int aVerbose=0);
  
  /**
   * Destructor of the TriggerBuffer class.
   */
  virtual ~TriggerBuffer(void);
  /**
     @}
  */

  
  /**
   * Flush the trigger buffer to the TTree.
   * The buffer is then reset.
   */
  bool FlushBuffer(void);

  /**
   * Resets the buffer.
   * All triggers saved until now are removed.
   */
  void ResetBuffer(void);

  /**
   * Defines new buffer time segments.
   * Before adding triggers to the buffer, time segments must be defined. This function defines the time segments associated to the buffer. In the following, triggers starting times must then belong to the buffer segments.
   *
   * Segments must be added in a chronological order and with no overlap with the previous segments. If this is not verified, false is returned. 
   * @param aSeg buffer segments
   */
  bool SetBufferSegments(Segments *aSeg);

  /**
   * Adds a trigger to the buffer.
   * Returns false if the buffer is full.
   *
   * This function overloads Triggers::AddTrigger() to fill the buffer instead the final structure. If the buffer size is 0, Triggers::AddTrigger() is called directly.
   * @param aTime trigger peak time
   * @param aFrequency trigger peak frequency
   * @param aSNR trigger SNR
   * @param aQ trigger Q
   * @param aTstart trigger time start
   * @param aTend trigger time end
   * @param aFstart trigger frequency start
   * @param aFend trigger frequency end
   * @param aAmplitude trigger amplitude
   * @param aPhase trigger phase
   */
  bool AddTrigger(const double aTime,   const double aFrequency,
		  const double aSNR,    const double aQ,
		  const double aTstart, const double aTend,
		  const double aFstart, const double aFend,
		  const double aAmplitude, const double aPhase);

  /**
   * Returns the time segments associated to a trigger selection.
   * This function only scans the triggers in the last segments added to the buffer (see SetBufferSegments()).
   * @param aThr threshold object
   * @param aInfValue value above which the threshold is considered infinite
   */
  Segments* GetLastTriggerSegments(TH1D *aThr=NULL, const double aInfValue=1e20);

  /**
   * Removes triggers (and segments) starting after a given time.
   * Entire segments are removed.
   * @param aTime time after which segments are removed
   */
  void ResetAfter(const double aTime);

  /**
   * Returns the size of the buffer.
   */
  inline int GetBufferSize(void){ return BufferSize; };

  /**
   * Returns the number of triggers currently in memory.
   * in the buffer or in the TTree.
   */
  inline int GetNtriggers(void){
    if(BufferSize) return b_pos;
    return MakeTriggers::GetNtriggers();
  };

  inline double GetBufferLiveTime(void){ return b_seg->GetLiveTime(); };

  
 private:
  
  // TRIGGERS
  int BufferSize;         ///< buffer size
  int b_pos;              ///< current buffer index
  double *b_time;         ///< trigger peak time
  double *b_freq;         ///< trigger peak frequency
  double *b_tstart;       ///< trigger time start
  double *b_tend;         ///< trigger time end
  double *b_fstart;       ///< trigger frequency start
  double *b_fend;         ///< trigger frequency end
  double *b_q;            ///< trigger Q
  double *b_snr;          ///< trigger SNR
  double *b_amp;          ///< trigger amplitude
  double *b_ph;           ///< trigger phase

  // SEGMENTS
  vector<double> s_start;    ///< list of start
  vector<double> s_end;      ///< list of end
  vector<int> s_pos;         ///< buffer index for that segment
  Segments *b_seg;           ///< buffer segments

  ClassDef(TriggerBuffer,0)  
};


#endif


